<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\Database\Booking;
use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Helpers\TimeHelper;

class BookingsStatisticsController extends RestController implements IRestfulController
{
    protected $route = 'booking-statistics';

    public function register_routes()
    {
        register_rest_route( $this->get_namespace(), '/' . $this->get_route () . '/totals', [            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_totals'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);

        register_rest_route( $this->get_namespace(), '/' . $this->get_route () . '/totals-by-date', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_totals_by_date'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);

        register_rest_route( $this->get_namespace(), '/' . $this->get_route () . '/all', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_all'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);

        register_rest_route( $this->get_namespace(), '/' . $this->get_route () . '/all/period', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_all_by_period'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);

        register_rest_route( $this->get_namespace(), '/' . $this->get_route () . '/period/total', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_period_total'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);
    }

    /**
     * @param \WP_REST_Request $request
     * @return \WP_REST_Response
     */
    function get_all(\WP_REST_Request $request){
        $page = $request->get_param('page');
        $per_page = $this->getPageLimitParam($request);

        if($page == null){
            $page = 1;
        }

        if($per_page == null){
            $per_page = 10;
        }

        $offset = ($page - 1) * $per_page;

        $result = array_map('self::transformSingle', Booking::getStatisticsAll($offset, $per_page));

        return $this->response( $result );
    }

    function get_all_by_period(\WP_REST_Request $request){
        $status = $request->get_param('status');
        $start_date = $request->get_param('start_date');
        $end_date = $request->get_param('end_date');
        $page = $request->get_param('page');
        $per_page = $this->getPageLimitParam($request);

        if($page == null){
            $page = 1;
        }

        if($per_page == null){
            $per_page = 10;
        }

        $offset = ($page - 1) * $per_page;

        $result = array_map('self::transformSingle', Booking::getStatisticsAllByPeriod($status, $start_date, $end_date, $offset, $per_page));

        return $this->response( $result );
    }

    function get_period_total(\WP_REST_Request $request){
        $status = $request->get_param('status' );
        $start_date = $request->get_param('start_date' );
        $end_date = $request->get_param('end_date' );

        return $this->response( Booking::getStatisticsPeriodTotal($status, $start_date, $end_date) );
    }

    /**
     * @param \WP_REST_Request $request
     * @return \WP_REST_Response
     */
    public function get_totals(\WP_REST_Request $request){
        $status = $request->get_param('status');

        $start_date = $request->get_param('start_date');
        $end_date = $request->get_param('end_date');

        return $this->response( Booking::getStatisticsTotals($status, $start_date, $end_date) );
    }

    /**
     * @param \WP_REST_Request $request
     * @return \WP_REST_Response
     */
    public function get_totals_by_date(\WP_REST_Request $request)
    {
        $status = $request->get_param('status');
        $date = $request->get_param('date');

        return $this->response( Booking::getStatisticsTotalsByDate($date, $status) );
    }

    /**
     * Transforms a single booking from database format to API format
     *
     * @param $booking
     * @return array
     */
    public static function transformSingle($booking)
    {
        $data = json_decode($booking['data'], true);

        return [
            'id' => $booking['id'],
            'title' => $booking['title'],
            'division_id' => self::getArrayValue($data, 'division_id'),
            'address' => self::getArrayValue($data, 'address.full_address'),
            'state' => self::getArrayValue($data, 'address.state'),
            'suburb' => self::getArrayValue($data, 'address.suburb'),
            'country' => self::getArrayValue($data, 'address.country'),
            'postcode' => self::getArrayValue($data, 'address.postcode'),
            'services' => self::getArrayValue($data, 'services'),
            'status' => $booking['status'],
            'error_code' => isset($booking['error_code']) ? $booking['error_code'] : null,
            'created_at' => date('d/m/Y H:i:s', strtotime($booking['created_at'])),
        ];
    }

    public static function getArrayValue($array, $key, $default = null){
        //if key contains dot, get the value from the nested array
        if(str_contains($key, '.')){
            $keys = explode('.', $key);
            $value = $array;

            foreach ($keys as $key){
                $value = self::getArrayValue($value, $key);
            }

            return $value;
        }

        return isset($array[$key]) ? $array[$key] : $default;
    }

    /**
     * @param \WP_REST_Request $request
     * @return mixed|null
     */
    public function getPageLimitParam(\WP_REST_Request $request): mixed
    {
        $per_page = $request->get_param('per_page', null);
        $page_limit = $request->get_param('page_limit', null);

        if($page_limit !== null){
            $per_page = $page_limit;
        }

        return $per_page;
    }

    /**
     * @param \WP_REST_Request $request
     * @param mixed $start_date
     * @param mixed $end_date
     * @return array
     */
    public function convertDates(\WP_REST_Request $request, mixed $start_date, mixed $end_date): array
    {
        if ($request->has_param('timezone')) {

            $timezone = $request->get_param('timezone');

            //replace - with /
            $timezone = str_replace('-', '/', $timezone);

            //check if timezone has more than 3 characters
            if (strlen($timezone) > 3) {
                $start_date = TimeHelper::convertToUTC($start_date, $timezone);
                $end_date = TimeHelper::convertToUTC($end_date, $timezone);
            } else {
                $start_date = TimeHelper::convertDateToTimezoneFromCountryCode($start_date, $timezone);
                $end_date = TimeHelper::convertDateToTimezoneFromCountryCode($end_date, $timezone);
            }
        }
        return array($start_date, $end_date);
    }

    /**
     * @param $request \WP_REST_Request
     * @return string
     */
    public function getTimezoneString($request){
        if ($request->has_param('timezone')) {
            $timezone = $request->get_param('timezone');
            //check if timezone has more than 3 characters
            if (strlen($timezone) > 3) {
                return $timezone;
            } else {
                return TimeHelper::getTimezoneFromCountryCode($timezone);
            }
        }
    }
}