<?php

namespace JimsCustomerBooking\RestfulAPI;

use JimsCustomerBooking\Admin\BackendPages\NativeFormSettingsPage;
use JimsCustomerBooking\BookingForm;
use JimsCustomerBooking\Database\Booking;
use JimsCustomerBooking\Generators\CBURLGenerator;
use JimsCustomerBooking\Helpers\Http;
use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\Integrations\ANotification;
use JimsCustomerBooking\Integrations\JimsFmsBooking;
use JimsCustomerBooking\Integrations\JimsLeadsIntegration;
use JimsCustomerBooking\Integrations\SlackNotification;
use JimsCustomerBooking\Jims_Plugin;
use JimsGroup\Apis\CustomerBooking;
use JimsGroup\Datatypes\BookingResponseType;
use JimsGroup\Datatypes\Datatype;
use JimsGroup\Facade\Authenticate;
use WpOrg\Requests\Exception;
use function categoryPosts\register_route;

class CustomerBookingController extends RestController implements IRestfulController
{
    protected $route = 'customer-booking';

    protected $findajim_autocomplete_token = Jims_Plugin::FINDAJIM_TOKEN;

    protected $fms_creds = Jims_Plugin::FMS_TOKENS;

    /**
     * @var string $findajim_api_mode
     */
    protected $findajim_api_mode = 'production'; // 'production', 'development

    /**
     * @var string[] $findajim_endpoints_base_url
     */
    protected $findajim_endpoints_base_url = [
        'development' => 'http://127.0.0.1:8001',
        'production' => 'https://findajim.jims.net',
    ];

    /**
     * @var string[] $findajim_endpoints
     */
    protected $findajim_endpoints = [
        'autocomplete' => '/api/address/autocomplete',
        'autocomplete_details' => '/api/address/autocomplete?with_details=1',
        'details' => '/api/address/details',
    ];

    protected $messages = [
        'success' => 'Success! Your booking is confirmed.',
        //error code 10 as default error code Booking unsuccessful
        'error' => 'Booking Processing Error. We apologize for the inconvenience. Unfortunately, we cannot process your request online at this time. However your inquiry has been forwarded to our Call Centre, and they will call you shortly.',
    ];


    private $available_modes = Jims_Plugin::FMS_API_MODES;

    private $error_messages = [
        'fms_unavailable' => 'Booking Processing Error. We apologize for the inconvenience. Unfortunately, we cannot process your request online at this time. However your inquiry has been forwarded to our Call Centre, and they will call you shortly.',
    ];

    private $mode = 'production'; // 'production', 'uat', 'hf', 'hf2', 'development

    public function register_routes()
    {
        register_rest_route($this->get_namespace(), $this->get_route() . '/cb-link', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_cb_link'),
                'permission_callback' => '__return_true',
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/list', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_bookings'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/create', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'create_booking'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/url/fse-photo', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_fse_photo'),
                'permission_callback' => '__return_true',
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/validate', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'validate_booking'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/single', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => array($this, 'get_booking'),
                'permission_callback' => [$this, 'checkPermissions'],
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/address', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'get_address')
            ]
        ]);

        register_rest_route($this->get_namespace(), $this->get_route() . '/address/details', [
            [
                'methods' => \WP_REST_Server::CREATABLE,
                'callback' => array($this, 'get_address_details'),
            ]
        ]);

        register_rest_route($this->get_namespace(), '/' . $this->get_route() . '/info', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_info'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);

        register_rest_route($this->get_namespace(), '/' . $this->get_route() . '/fms-auth-token', [
            [
                'methods' => \WP_REST_Server::READABLE,
                'callback' => [$this, 'get_fms_connect_token'],
                'permission_callback' => '__return_true',
                'args' => $this->get_collection_params()
            ]
        ]);
    }

    /**
     * @param \WP_REST_Request $request
     * @return \WP_REST_Response
     */
    public function get_info(\WP_REST_Request $request)
    {
        $decode_name = $request->get_param('decode');


        return $this->response([
            'version' => Jims_Plugin::$version,
            'division_id' => OptionHelper::getFormOption('bookingFormDivisionId'),
            'name' => !empty($decode_name) ? htmlspecialchars_decode(get_bloginfo('name')) : get_bloginfo('name'),
            'wordpressVersion' => get_bloginfo('version'),
            'phpVersion' => phpversion(),
            'debugInfo' => []
        ]);
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function get_fse_photo(\WP_REST_Request $request)
    {
        // /api/CutomerBookings/get-fse-photo/{fileId}/{iscontractor?}/{businessId?}
        /// /api/CustomerBookings/get-fse-photo/0/0/281610
        $is_contractor = 0;
        if ($request->has_param('is_contractor')) {
            $is_contractor = $request->get_param('is_contractor');
        }

        $fse_business_id = $request->get_param('fse_business_id');
        $fmsBooking = new JimsFmsBooking('AU', $this->getFMSAuthToken(), $this->getMode());
        $fmsBooking->setAuthHeaderToken( $this->getFMSAuthToken() );

        $path = [
            '/api/CustomerBookings/get-fse-photo/0/0/{{business_id}}',
//            '/api/CustomerBookings/get-fse-photo/0/1/{{business_id}}'
            '/Content/uploads/users/{{business_id}}/small.jpg',
        ][$is_contractor];


        $api_url = $fmsBooking->getApiUrl() . $path;

        //replace business_id
        $api_url = str_replace('{{business_id}}', $fse_business_id, $api_url);

        return $this->response([
            'status' => 'success',
            'url' => $api_url
        ]);
    }

    public function get_cb_link($request)
    {
        $properties = \JimsCustomerBooking\Jims_Plugin::getOptions()->getProperties();

        if (count($properties) == 0) {
            return $this->response([
                'status' => 'error',
                'message' => 'Please set up the plugin first'
            ]);
        }

        $cb_link = new CBURLGenerator($properties);
        return $this->response([
            'status' => 'success',
            'cb_link' => $cb_link->generate(false)
        ]);
    }


    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function create_booking($request)
    {
        $data = json_decode($request->get_body(), true);

        if ($request->has_param('booking_step') && $request->get_param('booking_step') == 1) {
            return $this->do_create_booking_allocation_step($data);
        }

        if ($request->has_param('booking_step') && $request->get_param('booking_step') == 2) {
            return $this->do_create_booking_final_step($data);
        }

        return $this->create_direct_booking($data);
    }

    public function get_fms_connect_token()
    {
        error_log('get_fms_connect_token', 0);
        return [
            'status' => 'success',
            'connect_token' => $this->getFMSAuthToken()
        ];
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function validate_booking(\WP_REST_Request $request)
    {
        $this->validate(json_decode($request->get_body(), true), $errors);
        return $this->response([
            'status' => count($errors) == 0 ? 'success' : 'error',
            'errors' => $errors
        ]);
    }


    /**
     * @param $data array
     * @return \WP_REST_Response
     */
    protected function do_create_booking_allocation_step($data)
    {
        $result = false;
        $fmsData = [];
        $notification_sent = false;
        $errors = [];

        if (!$this->validate($data, $errors)) {
            return $this->response(
                $this->getResponseArray('error', false, ['error' => $errors], false, $this->messages)
            );
        }


        //modify data['address']['full_address'], find suburb in full address and add post code after it
        $data['address']["full_address"] = $this->formatAddressForFMS($data['address']);


        //replace suburb for new zealand
        if ($data['address']['country'] === 'NZ') {
            $parts = explode(',', $data['address']["full_address"]);
            if (isset($parts[1])) {
                $data['address']["suburb"] = trim($parts[1]);
            }

            $data['address']["state"] = 'NZ';
            $data['address']['postcode'] = 'NZ ' . $data['address']['postcode'];

            //convert all characters in all values of address to nz special characters
            $enabled_nz_special_chars_replacement = OptionHelper::getFormOption('replace_nz_special', 0);
            if ($enabled_nz_special_chars_replacement == 1) {
                foreach ($data['address'] as $key => $value) {
                    $data['address'][$key] = is_string($value) ? StringHelper::convertSpecialNZSpecialChars($value) : $value;
                }
            }
        }

        //save lead
        list($booking, $booking_id, $data) = $this->saveLead($data);

        if ($booking_id === 0) {
            $messages = $this->messages;
            $messages['error'] = 'Quote request is completed. You cannot make this request again. Please select other services.';
            return $this->response(
                $this->getResponseArray('error', false, [], false, $messages)
            );
        }


        //Finalize booking to FMS
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getFMSAuthToken(),
                $this->getMode(),
            );


            //checks if the fms initial request hasn't return any errors
            if (count($FMSBooking->errors) > 0) {
                $this->updateBookingStatus($booking, $booking_id, false, 9);
                return $this->response(
                    $this->getResponseArray('error', false, [], false, [
                        'error' => $this->error_messages['fms_unavailable']
                    ])
                );
            }

            error_log(json_encode($data), 0);

            //extract data
            extract($this->formatBookingFormData($data));

            $booking_form = new BookingForm(
                $customer,
                $address,
                $division_id,
                $services
            );

            $response = $FMSBooking->doAllocationRequest($booking_form);
            $fmsData = $response->toArray();


            //check if booking was direct
            if ($response->isBookingCompleted()) {
                $result = $this->validateBookingResponse($response);
                if ($result) $this->updateBookingStatus($booking, $booking_id, true, $response->getPayloadProperty('bookingConfirmCode'));
            } else {
                //check if fmsData has error code
                if ($this->responseHasError($response)) {
                    $this->sendNotifications($booking_id, false, ['error_code' => $this->getErrorCode($response)]);
                    $this->updateBookingStatus($booking, $booking_id, false, $response->getPayloadProperty('bookingConfirmCode'));
                }

                //if not direct, respond to the frontend with the data
                return $this->response([
                    'status' => 'step',
                    'result' => isset($fmsData['result']) ? $fmsData['result'] : true,
                    'fms_data' => $fmsData,
                    'booking_id' => $booking_id,
                ]);
            }

        } catch (Exception $e) {
            //in case of error, respond to the frontend with the data
            $error_massage = $e->getMessage();
            $this->messages['error'] = empty($error_massage) ? $this->messages['error'] : $error_massage;
            $error_code = $this->getErrorCode($response);

            if (!empty($fmsData)) {
                $this->updateBookingStatus($booking, $booking_id, false, $error_code);
            } else {
                $this->updateBookingStatus($booking, $booking_id, false, 9);
            }


            $this->sendNotifications($booking_id, false, ['error_code' => $error_code]);
            return $this->response(
                $this->getResponseArray('error', false, $fmsData, false, $this->messages)
            );
        }

        // Send notification
        $notification_sent = $this->sendNotifications($booking_id, $result, ['error_code' => $response->getPayloadProperty('bookingConfirmCode')]);

        //if the booking is completed, update booking status and respond
        return $this->response(
            $this->getResponseArray($result ? 'success' : 'error', $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param $data array
     * @return \WP_REST_Response
     * @throws \JsonException
     */
    protected function do_create_booking_final_step($data): \WP_REST_Response
    {
        $result = false;
        $notification_sent = false;
        $fmsData = [];
        $errors = [];

        $this->validate($data, $errors);
        if (!isset($data['previous_response'])) {
            $errors['previous_response'] = true;
        }

        $valid = empty($errors);

        if (!$valid) {
            return $this->response(
                $this->getResponseArray('error', false, ['errors' => $errors], false, $this->messages)
            );
        }

        error_log( json_encode($data, JSON_THROW_ON_ERROR), 0 );

        //prepare for booking update
        $booking = new Booking();
        $booking_id = isset($data['booking_id']) ? $data['booking_id'] : null;

        extract($this->formatBookingFormData( $data ) );

        //Finalize booking to FMS
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getFMSAuthToken(),
                $this->getMode(),
            );
            $used_jims_before = isset( $data['used_jims_before'] ) ? $data['used_jims_before'] : false;
            $is_contractor = isset( $data['is_contractor'] ) ? $data['is_contractor'] : false;

            //remove is_contractor from data
            unset( $data['is_contractor'] );

            $booking_form = new BookingForm(
                $customer,
                $address,
                $division_id,
                $services
            );

            $PreviousResponse = new BookingResponseType(json_encode($data['previous_response']));

            error_log('Used Jims Before: ' . $used_jims_before, 0);
            $response = $FMSBooking->doFinalRequest($booking_form, $PreviousResponse, $used_jims_before, $is_contractor);
            $fmsData = $response->toArray();

            if ($response->isBookingCompleted()) {
                $result = $this->validateBookingResponse($response);
            }

            //check if fmsData has error code
            if ($this->responseHasError($response)) {
                $notification_sent = $this->sendNotifications($booking_id, false, ['error_code' => $response->getPayloadProperty('bookingConfirmCode')]);
                $this->updateBookingStatus($booking, $booking_id, false, $this->getErrorCode($response));
            }

        } catch (Exception $e) {
            $this->messages['error'] = $e->getMessage();

            //check if fmsData has error code
            if ($this->responseHasError($response)) {
                $this->updateBookingStatus($booking, $booking_id, false, $this->getErrorCode($response));
                $notification_sent = $this->sendNotifications($booking_id, false, ['error_code' => $this->getErrorCode($response)]);
            }

            error_log('error final step');
            return $this->response(
                $this->getResponseArray('error', false, $fmsData, $notification_sent, $this->messages)
            );
        } catch (\Exception $e) {
            //case of aunauthorized
            error_log('unauthorized final step');
        }

        if (!empty($fmsData) && isset($fmsData['payload'])) {
            $error_code = $fmsData['bookingConfirmCode'];;
            $this->messages['error'] = BookingResponseType::get_status_message((int) $error_code);
        }

        // Update booking status
        if (!empty($data['booking_id'])) {
            $this->updateBookingStatus($booking, $data['booking_id'], $result, $error_code);

            // Send notification
            $notification_sent = $this->sendNotifications($data['booking_id'], $result, ['error_code' => $response->getPayloadProperty('bookingConfirmCode')]);
        }

        error_log('final step normal response');
        return $this->response(
            $this->getResponseArray(
                $result ? 'success' : 'error',
                $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param array $data
     * @return \WP_REST_Response
     */
    protected function create_direct_booking(array $data)
    {
        $validation_errors = [];
        if (!$this->validate($data, $validation_errors)) {
            return $this->response(
                $this->getResponseArray('error', false, ['errors' => $validation_errors], false, $this->messages)
            );
        }

        list($booking, $booking_id, $data) = $this->saveLead($data);

        // Send to FMS
        list($FMSBooking, $result) = $this->connectToFMSAndSendData($data, $booking, $booking_id);

        $fmsData = $FMSBooking->getMessages(); //GET FMS RESPONSES
        list($result) = $this->updateMessages($fmsData, $result);

        // Send notification
        $notification_sent = $this->sendNotifications($booking_id, $result);

        //respond to the frontend with the data
        return $this->response(
            $this->getResponseArray($result ? 'success' : 'error', $result, $fmsData, $notification_sent, $this->messages)
        );
    }

    /**
     * @param $lastMessage
     * @return string
     */
    public function getLastMessage(array $fmsData)
    {
        //get last message
        $lastMessageData = $this->getTheLastNotEmptyItem($fmsData, true);

        $lastMessage = StringHelper::isJson($lastMessageData) ? json_decode($lastMessageData, true) : $lastMessageData;

        error_log(__FILE__ . ' ' . var_export($lastMessage, true), 0); //log last response as final response

        if (is_array($lastMessage)) {
            return $lastMessage['message'] ?? '';
        }

        return $lastMessage;
    }

    /**
     * @param $status
     * @param $result
     * @param $lastMessage
     * @param $notification_sent
     * @param $messages
     * @return array
     */
    protected function getResponseArray($status, $result, $lastMessage, $notification_sent, $messages)
    {
        if (!is_bool($result)) {
            $result = false;
        }

        return [
            'status' => $status,
            'result' => $result,
            'fms_data' => $lastMessage,
            'notification' => $notification_sent ? 'sent' : 'not sent',
            'notification_message' => $result ? $messages['success'] : $messages['error'],
        ];
    }

    /**
     * @param $array
     * @return array|null
     */
    protected function getTheLastNotEmptyItem($array, $onlyArrays = true)
    {
        $last = null;
        foreach ($array as $item) {

            $array_result = true;
            if ($onlyArrays) {
                $array_result = is_array($item);
            }

            if ($array_result && !empty($item)) {
                $last = $item;
            }
        }

        return $last;
    }

    /**
     * @param $request
     * @return \WP_REST_Response
     */
    public function get_bookings($request)
    {
        $page = $limit = $status = null;
        $values = $this->getSearchParams($request);

        extract($values);

        //create transient key
        $transient_key = StringHelper::GetPrefixedFromParts(['customer_booking_list', $page, $limit, $status], 'machine');


        if (false === ($result = get_transient($transient_key))) {
            $booking = new Booking();
            $data = $booking->getPage($page, $limit, $status);

            $result = array_map(function ($item) {
                return [
                    'id' => $item['booking_id'],
                    'data' => json_decode($item['data']),
                    'status' => $item['booking_status'],
                    'created_at' => $item['created_at'],
                    'updated_at' => $item['updated_at']
                ];
            }, $data);

            set_transient($transient_key, $result, 0.5 * HOUR_IN_SECONDS);
        }

        return $this->response([
            'status' => 'success',
            'page_count' => Booking::getPageCount($limit, $status),
            'current_page' => $page,
            'data' => $result
        ]);
    }

    /**
     * @param $request \WP_REST_Request
     * @return \WP_REST_Response
     */
    public function get_booking($request)
    {
        $booking = new Booking();

        $id = $request->get_param('booking_id');
        $data = $booking->getFromId($id);

        return $this->response($data);
    }

    /**
     * @param $request
     * @return array|int[]
     */
    protected function getSearchParams($request): array
    {
        $params_defaults = [
            'page' => 1,
            'limit' => 10,
            'status' => 'all'
        ];

        $data = $request->get_params();
        $search_params = \array_keys($params_defaults);

        $values = \array_combine($search_params, \array_map(function ($key) use ($data, $params_defaults) {
            return isset($data[$key]) ? $data[$key] : $params_defaults[$key];
        }, $search_params));

        return $values;
    }

    /**
     * @param $request \WP_REST_Request
     * @return mixed
     */
    public function get_address($request)
    {
        $address = $request->get_param('address');
        $country_code = $request->get_param('country_code');

        $key = StringHelper::GetPrefixedFromParts(['booking', 'address', $address], 'machine');

        $response = $this->getResponse($key, [
            'address' => $address,
            'country_code' => $country_code
        ], 0.5 * HOUR_IN_SECONDS, 'get_address_suggestions');

        return json_decode($response['result']);
    }

    public function get_address_suggestions($address, $country_code = null)
    {
        return Http::post($this->getFindAJimEndpoint('autocomplete'), [
            'address' => $address,
            'countryCode' => empty($country_code) ? 'AU' : $country_code,
            'token' => $this->findajim_autocomplete_token
        ]);
    }

    /**
     * @param $request
     * @return mixed
     */
    public function get_address_details($request)
    {
        $address = $request->get_param('address');

        $key = StringHelper::GetPrefixedFromParts(['booking', 'address', 'details', $address], 'machine');

        if (false === ($response = get_transient($key))) {

            $response = Http::post($this->getFindAJimEndpoint('details'), [
                'address' => $address,
                'token' => $this->findajim_autocomplete_token
            ]);

            set_transient($key, $response, 0.5 * HOUR_IN_SECONDS);
        }

        return json_decode($response['result']);
    }

    /**
     * @param mixed $data
     * @param Booking $booking
     * @param int $booking_id
     * @return array
     */
    protected function connectToFMSAndSendData(mixed $data, Booking $booking, int $booking_id): array
    {
        $result = false;
        //Connect to FMS and send data
        try {
            $FMSBooking = new JimsFmsBooking(
                $data['address']['country'],
                $this->getFMSAuthToken(),
                $this->getMode(),
            );
            $FMSBooking->setAuthHeaderToken( $this->getFMSAuthToken() );

            error_log(__FILE__ . ' ' . json_encode($data), 0);

            extract($this->formatBookingFormData($data));

            $result = $FMSBooking->doBookingRequest(
                $customer,
                $address,
                $division_id,
                $services
            );
        } catch (\Exception $e) {
            error_log(__FILE__ . ' ' . $e->getMessage(), 0);
            $result = false;
        }

        $this->updateBookingStatus($booking, $booking_id, $result);

        return array($FMSBooking, $result);
    }

    /**
     * @return string
     */
    public function getMode()
    {
        $settings = OptionHelper::get(NativeFormSettingsPage::$option_name, []);

        if (isset($settings['fms_environment']) && in_array($settings['fms_environment'], $this->available_modes)) {
            return $settings['fms_environment'];
        }

        return $this->mode;
    }

    /**
     * @param $fmsData
     * @param bool $result
     * @return array
     */
    public function updateMessages($fmsData, bool $result): array
    {
        if (count($fmsData) > 0) {

            $fmsData = array_map(function ($item) {
                return StringHelper::isJson($item) ? json_decode($item, true) : $item;
            }, $fmsData);

            $lastMessage = $this->getLastMessage($fmsData);

            if (strtolower($lastMessage) !== 'booked') {
                $result = false;
            }

            $this->messages['error'] = $lastMessage ?? $this->messages['error'];
        }
        return array($result);
    }

    /**
     * @param BookingResponseType $response
     * @param mixed $booking
     * @param mixed $booking_id
     * @return true
     * @throws Exception
     */
    public function validateBookingResponse(BookingResponseType $response): bool
    {
        if ($response->getPayloadProperty('bookingConfirmCode') === BookingResponseType::CONFIRM_CODE_BOOKED) {
            error_log('Booking completed', 0);
        } else {
            error_log($response->getStatusMessage());
            throw new Exception($response->getStatusMessage(), 0);
        }

        return true;
    }

    /**
     * @param $response
     * @return bool
     */
    public function responseHasError($response)
    {
        $responseCode = (int)$response->getPayloadProperty('bookingConfirmCode');
        return $responseCode > 0 && $responseCode <= 9;
    }


    /**
     * @param array $address
     * @return string
     */
    protected function formatAddressForFMS(array $address): string
    {
        //format the address
        //{{unit number}} {{street number}} {{street name}}, {{suburb}}, {{state}}, {{postcode}}, {{country}}
        $address = sprintf('%s%s %s, %s, %s, %s, %s',
            !empty($address['unit_number']) ? $address['unit_number'] . '/' : '',
            $address['street_number'],
            $address['street_name'],
            $address['suburb'],
            $address['state'],
            $address['postcode'],
            $address['country']
        );

        //remove space in the start
        $address = ltrim($address);

        //remove space in the end
        $address = rtrim($address);

        //remove double spaces
        return preg_replace('/\s+/', ' ', $address);
    }

    /**
     * @param BookingResponseType $response
     * @return int|mixed|null
     */
    public function getErrorCode(BookingResponseType $response, $default = 9): mixed
    {
        return !empty($response->getPayloadProperty('bookingConfirmCode')) ? $response->getPayloadProperty('bookingConfirmCode') : $default;
    }

    /**
     * @param $data
     * @return BookingForm
     */
    protected function getBookingForm($data)
    {
        list($customer, $address, $division_id, $service_ids) = $this->formatBookingFormData($data);
        return new BookingForm($customer, $address, $division_id, $service_ids);
    }

    /**
     * @param $data
     * @return array
     */
    protected function formatBookingFormData($data): array
    {
        return [
            'customer' => [
                'address' => null,
                'addressString' => $data['address']['full_address'],
                'allow_rating' => true,
                'comments' => $data['comments'],
                'isUserInfoConfirmed' => null,
                'emailAddress' => $data['email'],
                'firstName' => $data['firstname'],
                'lastName' => $data['lastname'],
                'phoneNumber' => $data['phone'],
                'subscribed' => true,
                'businessName' => null
            ],
            'address' => $data['address'],
            'division_id' => $data['division_id'],
            'services' => $data['services'],
        ];
    }

    /**
     * @param int $booking_id
     * @param mixed $result
     * @return bool
     */
    protected function sendNotifications(int $booking_id, mixed $result, $data = []): bool
    {
        $notification_sent = $this->sendSlackNotification($booking_id, $result, $data);

        $error_code = $this->getErrorCodeForNotifications($booking_id, $data['error_code'] ?? null);

        //initialize Leads Integration
        $leads_integration = new JimsLeadsIntegration();

        //send notification to leads integration
        $leads_integration->updateRealTimeBooking($booking_id, $error_code);

        //check if error code is 5,6,9 or 10
        if (in_array($error_code, [5, 9, 10])) {
            $leads_integration->sendErrorNotificationEmail($booking_id, $error_code);
        }

        return $notification_sent;
    }

    protected function sendSlackNotification(int $booking_id, mixed $result, $data = [])
    {
        $notification = new SlackNotification('https://hooks.slack.com/services/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', [
            'error_id' => sprintf('%s', "{$booking_id}"),
            'booking_id' => sprintf('%s', "{$booking_id}"),
            'website' => get_site_url(),
            'error_code' => $data['error_code'] ?? ''
        ]);

        $notification_sent = $notification->send($result ? ANotification::NOTIFICATION_TYPE_INFO : ANotification::NOTIFICATION_TYPE_ERROR);
        return $notification_sent;
    }

    /**
     * @param $endpoint_name
     * @return string
     */
    protected function getFindAJimEndpoint($endpoint_name): string
    {
        return sprintf('%s%s',
            $this->findajim_endpoints_base_url[$this->findajim_api_mode],
            $this->findajim_endpoints[$endpoint_name]
        );
    }

    /**
     * @param string $key
     * @param array $params
     * @param int $cache_ttl
     * @param string $callback_name
     * @return array|mixed
     */
    protected function getResponse(string $cache_key, array $params, $cache_ttl, $callback_name): mixed
    {

        $cached_response = $response = get_transient($cache_key);
        if (false === ($cached_response)) {
//            $response = $this->get_address_suggestions ($address, $country_code);
            $response = call_user_func_array([$this, $callback_name], $params);
            set_transient($cache_key, $response, $cache_ttl);
        }

        if ($this->stored_response_is_empty($cached_response)) {
            $response = call_user_func_array([$this, $callback_name], $params);
        }
        return $response;
    }

    protected function validate(array $data, &$errors = [])
    {
        $required_fields = [
            'firstname',
            'lastname',
            'email',
            'phone',
            'address',
            'division_id',
            'services',
            'comments'
        ];

        foreach ($required_fields as $field) {
            if ($field === 'comments' && isset($data[$field])) {
                continue;
            }

            if (empty($data[$field])) {
                $errors[$field] = true;
            }
        }

        if (empty($errors) && !$this->validateAddressObject($data['address'])) {
            $errors['address'] = true;
        }

        return empty($errors);
    }

    protected function validateAddressObject(array $address)
    {
        $required_fields = [
            'full_address',
//            'country',
//            'state',
//            'suburb',
//            'postcode',
            'street_number',
            'street_name',
            'geocode_y',
            'geocode_x',
        ];

        $is_valid = true;
        foreach ($required_fields as $field) {
            if (empty($address[$field])) {
                $is_valid = false;
                break;
            }
        }
        return $is_valid;
    }

    /**
     * @param array $data
     * @return array
     */
    protected function saveLead(array $data, $error_code = 10): array
    {
        $booking = new Booking();
        $booking_id = $booking->add($data, 'pending', $error_code);

        if (isset($data['address'])) {
            unset($data['address']['place_id']);
//            unset($data['address']['unit_number']);
        }
        return array($booking, $booking_id, $data);
    }

    /**
     * @param Booking $booking
     * @param int $booking_id
     * @param bool $result
     * @return void
     */
    protected function updateBookingStatus(Booking $booking, int $booking_id, bool $result = false, $error_code = 10): void
    {
        if (!empty($booking_id))
            $booking->updateStatus($booking_id, $result ? 'sended' : 'pending', $error_code);
    }

    /**
     * @param $booking_id int
     * @param $error_code int|null
     * @return int
     */
    protected function getErrorCodeForNotifications($booking_id, $error_code)
    {
        if ($error_code === 0 || $error_code === '0') {
            return 0;
        }

        $error_code = (int)$error_code;

        if (empty($error_code)) {
            $booking = new Booking();
            $booking_data = $booking->getFromId($booking_id, Booking::RESULT_TYPE_ARRAY_A);

            if ($booking_data !== null) {
                $error_code = (int)$booking_data['error_code'];
            } else {
                $error_code = 10;
            }
        }

        return $error_code;
    }

    protected function getFMSAuthToken(): string
    {
        $transient_key = StringHelper::GetPrefixedFromParts(['fms', 'auth_token'], 'machine');
        if (false === ($auth_token = get_transient($transient_key))) {

            $auth = new Authenticate();
            $auth->setFmsConfig($this->fms_creds);
            [$data, $auth_token] = $auth->CustomerBookingAccessToken($this->getMode());


            set_transient($transient_key, $auth_token, 0.5 * HOUR_IN_SECONDS);
        }

        return $auth_token;
    }
}