<?php

namespace JimsCustomerBooking\Services;

use JimsCustomerBooking\Admin\BackendPages\NativeFormSettingsPage;
use JimsCustomerBooking\Helpers\Http;
use JimsCustomerBooking\Helpers\OptionHelper;
use JimsCustomerBooking\Helpers\StringHelper;
use JimsCustomerBooking\Jims_Plugin;
use JimsCustomerBooking\Validators\JsonValidator;
use JimsGroup\Api;
use JimsGroup\Apis\CustomerBooking;
use JimsGroup\Facade\Authenticate;

if (!class_exists('\JimsCustomerBooking\Services\DivisionService')) {
    /**
     * TODO: Update this to use FMS API instead of internal.jims.net
     */
    class DivisionService extends Service
    {

        private $divisionsLink = JIMS_INTERNAL_URL . '/api/division/all';

        private $countriesLink = JIMS_INTERNAL_URL . '/api/division/countries/active/all';

        protected $default_countries = [
            [
                'name' => 'Australia',
                'symbol' => 'AU'
            ],
            [
                'name' => 'New Zealand',
                'symbol' => 'NZ'
            ],
            [
                'name' => 'United Kingdom',
                'symbol' => 'UK'
            ],
            [
                'name' => 'Canada',
                'symbol' => 'CA'
            ]
        ];

        /**
         * @var Api|null $Api
         */
        private $Api = null;

        public function __construct()
        {
            $this->Api = new Api();
        }

        /**
         * @return array|mixed
         * @throws \Exception
         */
        public function call(): mixed
        {
            switch ($_GET['part']) {
                case 'divisions':
                    return $this->getDivisions();
                case 'services':
                    return $this->getServices();
            }
        }

        /**
         * @param $link
         * @return mixed
         */
        public function getDivisions($short = true): mixed
        {
            $output = [];
            if (false === ($value = get_transient(StringHelper::GetPrefixed('divisions')))) {
                $customDivisions = $this->getCustomDivisions();

                if (empty($customDivisions)) {
                    $fmsInfo = $this->Api->getFMSInfoApi();
                    $response = $fmsInfo->getDivisions('json');
                } else {
                    $response = $customDivisions;
                }

                set_transient(StringHelper::GetPrefixed('divisions'), $response, 5 * DAY_IN_SECONDS);
                $output = json_decode($response);
            } else {
                $output = json_decode($value);
            }

            $icons = $this->getDivisionIcons();

            $short_output = [];
            if ($short && !empty($output)) {
                foreach ($output as $division) {
                    if (empty($division->name)) continue;
                    $short_output[] = [
                        'id' => $division->id,
                        'name' => strtoupper($division->name),
                        'icon' => !empty($icons[$division->id]['icon']) ? $icons[$division->id]['icon'] : JIMS_CBP_URL . 'assets/dist/img/jimsgroup.png',
                    ];
                }

                return $short_output;
            }

            return $output;
        }

        /**
         * @param $division_id
         * @return array|null
         */
        public function getDivisionById($division_id): array|null
        {
            $divisions = $this->getDivisions(true);
            $division = array_filter($divisions, function ($item) use ($division_id) {
                return $item->id == $division_id;
            });

            return array_shift($division);
        }

        /**
         * @return array|mixed
         */
        public function getServices($format = false, $countryCode = 'AU'): mixed
        {
            $transientName = StringHelper::GetPrefixed('services_' . $countryCode);

            if (false === ( $output = get_transient($transientName) ) ) {

                $customerBooking = new CustomerBooking();
                $customerBooking->setAuthToken( Jims_Plugin::FMS_API_KEY );
                $customerBooking->setHeaderAuthToken(
                    $this->getFMSAuthToken()
                );

                try {
                    // get initial booking request
                    $output = $customerBooking->getInitialBookingRequest([
                        'countryCode' => $countryCode,
                    ]);

                    // set transient
                    set_transient($transientName, $output, 1 * HOUR_IN_SECONDS);
                } catch (\Exception $e) {
                    throw new \Exception($e->getMessage());
                }
            }

            return $format === true ? $this->formatServices($output) : $output;
        }

        /**
         * @param string $return_type , object|array
         * @return array|mixed
         * @throws \JsonException
         */
        public function getCountries(string $return_type = 'object'): array
        {
            if (false === ($result = get_transient(StringHelper::GetPrefixed('countries')))) {
                [$result, $headers, $error] = Http::Get($this->countriesLink);

                //if there is an error, return the default countries
                if (!StringHelper::isJson($result) || empty($result)) {
                    $result = json_encode($this->default_countries, JSON_THROW_ON_ERROR);
                }

                set_transient(StringHelper::GetPrefixed('countries'), $result, 1 * DAY_IN_SECONDS);
            }

            return json_decode($result, $return_type === 'array', 512, JSON_THROW_ON_ERROR);
        }

        /**
         * @param $services
         * @return array|array[]
         */
        public function formatServices($services): array
        {

            return array_map(function ($service) {
                return [
                    'name' => $service->friendly_name,
                    'id' => $service->id,
                    'division_id' => $service->group_id
                ];
            }, $services);
        }


        protected function getCustomDivisions()
        {
            if (!Http::makeTestRequest($this->divisionsLink)) {
                return null;
            }

            list($result, $headers, $error) = Http::Get($this->divisionsLink);

            if ($error) return null;

            return $result;
        }

        /**
         * @return array|array[]|mixed
         */
        public function getDivisionIcons(): mixed
        {
            $fmsInfo = $this->Api->getFMSInfoApi();
            $auth = $this->getAuth();
            $fmsInfo->setHeaderAuthToken( $auth->CustomerBookingAccessToken()[1] );

            $transientName = StringHelper::GetPrefixed('fms_division_icons');
            if (false === ($requestResponse = get_transient($transientName))) {
                $requestResponse = $fmsInfo->getDivisionIcons();

                uasort($requestResponse, static function($a, $b){
                    return $a['id'] - $b['id'];
                });

                //store divisions in transient for 12 hours
                set_transient($transientName, $requestResponse, 12 * HOUR_IN_SECONDS);
            }

            return $requestResponse;
        }

        protected function getAuth(): \JimsGroup\Facade\Authenticate
        {
            $auth = new \JimsGroup\Facade\Authenticate;
            $auth->setFmsConfig(Jims_Plugin::FMS_TOKENS);

            return $auth;
        }

        protected function getFMSAuthToken(){
            $auth = new Authenticate();
            $auth->setFmsConfig(Jims_Plugin::FMS_TOKENS );
            [$data, $auth_token] = $auth->CustomerBookingAccessToken($this->getMode());

            return $auth_token;
        }

        protected function getMode($default = 'production')
        {
            $settings = OptionHelper::get(NativeFormSettingsPage::$option_name, []);

            if (isset($settings['fms_environment']) && in_array($settings['fms_environment'], Jims_Plugin::FMS_API_MODES)) {
                return $settings['fms_environment'];
            }

            return $default;
        }
    }
}
